/*
 *  Copyright: Collin Mulliner <collin[at]mulliner.org>
 *  Web: http://www.mulliner.org/nfc/
 *  
 *  License: GPLv2
 *  
 */

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Date;
import java.util.Vector;

import javax.microedition.contactless.ContactlessException;
import javax.microedition.contactless.DiscoveryManager;
import javax.microedition.contactless.TargetListener;
import javax.microedition.contactless.TargetProperties;
import javax.microedition.contactless.TargetType;
import javax.microedition.contactless.ndef.NDEFMessage;
import javax.microedition.contactless.ndef.NDEFRecord;
import javax.microedition.contactless.ndef.NDEFTagConnection;
import javax.microedition.io.Connector;
import javax.microedition.io.StreamConnection;
import javax.microedition.io.StreamConnectionNotifier;
import javax.microedition.lcdui.Command;
import javax.microedition.lcdui.CommandListener;
import javax.microedition.lcdui.Display;
import javax.microedition.lcdui.Displayable;
import javax.microedition.lcdui.Form;
import javax.microedition.lcdui.StringItem;
import javax.microedition.lcdui.TextField;
import javax.microedition.midlet.MIDlet;
import javax.microedition.midlet.MIDletStateChangeException;
import javax.microedition.rms.RecordEnumeration;
import javax.microedition.rms.RecordStore;

public class BtNfcAdapter extends MIDlet implements CommandListener, Runnable, TargetListener, DebugInterface
{
	private DiscoveryManager dm;
	private String uuid;
	private StreamConnectionNotifier btConNotifier;
	private StreamConnection clientCon;
	private DataInputStream clientInput;
	private DataOutputStream clientOutput;
	private boolean appRunning;
	
	private Thread runner;
	
	private Command exitcmd;
	private Command clearcmd;
	private Command writeextcmd;
	private Command writeintcmd;
	
	private Form mainScreen;
	private TextField numReadTags;
	private TextField sizeWriteTag;
	//private TextField clientAddr;
	private TextField status;
	private TextField cardType;
	
	private int write;
	//private int writeInternal;
	private Vector readTags;
	private int numTags;
	
	private NDEFMessage writeNdefMsg;
	private NFCTagCmd curCmd;
	private NFCTag writeTag;

	private RecordStore dataStore;
	
	private String dbg[];
	private int dbg_num;
	private int debug_level = 0;
	
	protected void destroyApp(boolean arg0) throws MIDletStateChangeException
	{
		appRunning = false;
		try {
			if (runner != null)
				runner.join();
		} catch (InterruptedException e) {}
	}
	
	public void debug(String d, int level)
	{
		StringBuffer ob = new StringBuffer();
		if (level == debug_level) {
			for (int i = 0; i < dbg_num - 1; i++) {
				dbg[i] = dbg[i+1];
			}
			dbg[dbg_num - 1] = d;
			for (int i = 0; i < dbg_num; i++) {
				if (dbg[i] != null) {
					ob.append(dbg[i]);
				}
			}
			status.setString(ob.toString());
		}
	}
	
	private void save()
	{
		try {
			RecordStore.deleteRecordStore("Tags");
		} catch (Exception e) {}
		try {
			dataStore = RecordStore.openRecordStore("Tags", true);
			for (int i = 0; i < numTags; i++) {
				NFCTag t = (NFCTag) readTags.elementAt((numTags - 1) - i);
				dataStore.addRecord(t.toRMS(), 0, t.toRMS().length);
				debug("saved tag", 0);
			}
			dataStore.closeRecordStore();
		} catch (Exception e) {}
	}

	protected void pauseApp() {}

	protected void startApp() throws MIDletStateChangeException
	{	
		exitcmd = new Command("Exit", Command.EXIT, 1);
		clearcmd = new Command("ClearTags", Command.ITEM, 1);
		writeextcmd = new Command("Write to External Tag", Command.ITEM, 1);
		writeintcmd = new Command("Write to Internal Tag", Command.ITEM, 1);
		mainScreen = new Form("Bluetooth NFC Adapter");
		mainScreen.addCommand(exitcmd);
		mainScreen.addCommand(clearcmd);
		mainScreen.addCommand(writeextcmd);
		//mainScreen.addCommand(writeintcmd);
		
		mainScreen.append(new StringItem("Version:", "0.2.7"));
		mainScreen.append(new StringItem("Author:", "Collin Mulliner <collin@mulliner.org>"));
		mainScreen.append(new StringItem("Web:", "http://www.mulliner.org/nfc/"));
		mainScreen.append(new StringItem("License:", "GPLv2"));
		
		numReadTags = new TextField("No. Tags", "0", 3, TextField.UNEDITABLE);
		sizeWriteTag = new TextField("Write Size", "0", 5, TextField.UNEDITABLE);
		//clientAddr = new TextField("Client", "", 20, TextField.UNEDITABLE);
		status = new TextField("Status", "started", 1024, TextField.UNEDITABLE);
		cardType = new TextField("Card Type:", "", 50, TextField.UNEDITABLE);
		
		//mainScreen.append(clientAddr);
		mainScreen.append(numReadTags);
		mainScreen.append(sizeWriteTag);
		mainScreen.append(cardType);
		mainScreen.append(status);
		
		Display.getDisplay(this).setCurrent(mainScreen);
	
		mainScreen.setCommandListener(this);
		registerTargetListner();
		
		uuid = new String("AABBCCDDAABBCCDDAABBCCDDAABBCCDD");
		write = 0;
		//writeInternal = 0;
		appRunning = true;
		readTags = new Vector();
		numTags = 0;
		
		dbg_num = 5;
		dbg = new String[dbg_num];
		
		try {
			dataStore = RecordStore.openRecordStore("Tags", true);
			RecordEnumeration re = dataStore.enumerateRecords(null, null, false);
			while (re.hasNextElement()) {
				byte[] b = re.nextRecord();
				NFCTag t = new NFCTag();
				t.fromRMS(b);
				readTags.addElement(t);
				numTags++;
			}
			dataStore.closeRecordStore();
		} catch (Exception e) {}
		numReadTags.setString(String.valueOf(numTags));
		
		runner = new Thread(this);
		runner.start();
	}

	public void commandAction(Command cmd, Displayable disp)
	{
		if (cmd == exitcmd) {
			save();
			notifyDestroyed();
		}
		else if (cmd == clearcmd) {
			readTags.removeAllElements();
			numTags = 0;
			numReadTags.setString(String.valueOf(numTags));
			debug("tag cache cleared", 0);
		}
		else if (cmd == writeextcmd) {
			if (numTags > 0) {
				writeTag = (NFCTag) readTags.elementAt(0);
				sizeWriteTag.setString(String.valueOf(writeTag.getHex().length));
				writeNdefMsg = writeTag.message;
				write = 1;
				readTags.removeElementAt(0);
				numTags--;
				numReadTags.setString(String.valueOf(numTags));
			}
		}
		else if (cmd == writeintcmd) {
			// doesn't work yet
			writeTag = (NFCTag) readTags.elementAt(0);
			writeNdefMsg = writeTag.message;
			readTags.removeElementAt(0);
			numTags--;
			numReadTags.setString(String.valueOf(numTags));
			debug("write internal tag ", 0);
			try {
				String uri = System.getProperty("internal.mf.url");
				debug("uri=" + uri + " ", 0);
				NDEFTagConnection conn = (NDEFTagConnection) Connector.open(uri);
				if (conn != null) {
					conn.writeNDEF(writeNdefMsg);
					conn.close();
				}
				else
					debug("internal connection failed ", 0);
			} catch (Exception e) {
				debug("int write err: " + e.getMessage() + " ", 0);
			}
			writeNdefMsg = null;
		}
	}
	
	private boolean waitForClient()
	{
		try {
			clientCon = btConNotifier.acceptAndOpen();
			clientInput = new DataInputStream(clientCon.openInputStream());
			clientOutput = new DataOutputStream(clientCon.openOutputStream());
		}
		catch (IOException e) {
			return false;
		}
		debug("client connected ", 0);
		return true;
	}
	
	private boolean sendNull()
	{
		String outp = new String("0\n");
		try {
			clientOutput.write(outp.getBytes());
			clientOutput.flush();
			return true;
		} catch (IOException e) {
			return false;
		}
	}
	
	public void run()
	{
		String server_url = new String("btspp://localhost:" + uuid + ";name=BtNfcAdapter");
		CommandInterface ci = null;
	
		try {
			btConNotifier = (StreamConnectionNotifier) Connector.open(server_url);
		} catch (IOException e) {}
		
		boolean closeCon = false;
		while (appRunning) {
			if (clientCon == null) {
				if (waitForClient())
					ci = new CommandInterface(clientInput, this);
				else
					clientCon = null;
			}
			else {
				debug("wait for cmd ", 0);
				curCmd = ci.getCommand();
				if (curCmd == null)
					closeCon = true;
				else {
					// read
					if (curCmd.getType() == 1) {
						if (readTags.isEmpty()) {
							debug("got Read Command, but no Tags read so far ", 0);
							if (!sendNull())
								closeCon = true;
						}
						else {
							debug("got Read Command ", 0);
							NFCTag tag = (NFCTag) readTags.elementAt(0);
							if (tag != null) {
								readTags.removeElementAt(0);
								numTags--;
								try {
									String outp = tag.toReadTagCmd();
									clientOutput.write(outp.getBytes());
									clientOutput.write(tag.getHex());
									clientOutput.write(new String("\n").getBytes());
									clientOutput.flush();
								} catch (IOException e) {
									debug(" error sending tag data to client: " + e.toString(), 0);
									closeCon = true;
								}
							}
							else {
								debug("got Read Command, tag is null ", 0);
								if (!sendNull())
									closeCon = true;
							}
							numReadTags.setString(String.valueOf(numTags));
						}
					}
					// write
					else if (curCmd.getType() == 2) {
						debug("got Write Command ", 0);
						writeTag = new NFCTag(curCmd.getData(), this);
						sizeWriteTag.setString(String.valueOf(curCmd.getData().length()));
						writeNdefMsg = writeTag.message;
						write = 1;
						if (writeNdefMsg == null) {
							debug("writeNdefMsg = null ", 0);
							write = 0;
							sizeWriteTag.setString("0");
						}
					}
				}
			}
			if (closeCon) {
				try {
					debug("client disconnected ", 0);
					clientInput.close();
					clientOutput.close();
					clientCon.close();
				} catch (IOException e) {}
				closeCon = false;
				clientCon = null;
			}
			// cleanup memory
			System.gc();
		}
	}
	
	private void targetDetected_read(TargetProperties[] prop)
	{
		NFCTag curTag = new NFCTag();
		
		curTag.mapping = prop[0].getMapping();
		curTag.uid = prop[0].getUid();
		curTag.url = prop[0].getUrl();
		curTag.targetTypes = prop[0].getTargetTypes();
		curTag.dateTime = new Date();
		
		Class[] tempcn = prop[0].getConnectionNames();
		curTag.connNames = new String[tempcn.length];
		for (int i = 0; i < tempcn.length; i++) {
			curTag.connNames[i] = tempcn[i].toString();
		}
		
		cardType.setString(curTag.getCardType(this));
		
		try {
			NDEFTagConnection nconn = (NDEFTagConnection) Connector.open(prop[0].getUrl());
			curTag.message = nconn.readNDEF();
		
			if (curTag.message != null) {
				NDEFRecord[] records = curTag.message.getRecords();
				curTag.records = new NDEFRecord[records.length];
			}
			nconn.close();
		}
		catch (Exception e) {
			debug("read err: " + e.getMessage(), 0);
			curTag.state = -1;
		}
		
		if (curTag.state != -1) {
			readTags.addElement(curTag);
			numTags++;
			debug("tag read ", 0);
			numReadTags.setString(String.valueOf(numTags));
		}
	}
	
	private void targetDetected_write(TargetProperties[] prop)
	{
		NDEFTagConnection conn = null;
		if (writeNdefMsg == null)
			return;
		
		try {
			conn = (NDEFTagConnection) Connector.open(prop[0].getUrl());
			conn.writeNDEF(writeNdefMsg);
        }
		catch (ContactlessException e) {
			debug("write err: " + e.toString(), 0);
        }
		catch (IOException e) {
			debug("write err: " + e.toString(), 0);
        }
		finally {
			if (conn != null) {
				try {
					conn.close();
				} catch (IOException e) {
					debug("write err: " + e.toString(), 0);
                }
            }
		}
		writeNdefMsg = null;
		debug("tag written ", 0);
		sizeWriteTag.setString("0");
	}
	
	public void targetDetected(TargetProperties[] prop)
	{
		if (write == 0) {
			targetDetected_read(prop);
		}
		else {
			targetDetected_write(prop);
			write = 0;
		}	
	}
	
	private void registerTargetListner()
	{
		dm = DiscoveryManager.getInstance();
		try {
			dm.addTargetListener(this, TargetType.NDEF_TAG);
			dm.addTargetListener(this, TargetType.ISO14443_CARD);
			dm.addTargetListener(this, TargetType.RFID_TAG);
		}
		catch (IllegalStateException e) {}
		catch (ContactlessException e) {}
	}
}